# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cython: language_level=3
# cython: binding=True

cdef extern from "<isa-l/igzip_lib.h>":
    # Deflate compression standard defines
    int ISAL_DEF_MAX_HDR_SIZE
    int ISAL_DEF_MAX_CODE_LEN
    int IGZIP_K
    int ISAL_DEF_HIST_SIZE
    int ISAL_DEF_MAX_HIST_BITS
    int ISAL_DEF_MAX_MATCH
    int ISAL_DEF_MIN_MATCH

    int ISAL_DEF_LIT_SYMBOLS
    int ISAL_DEF_LEN_SYMBOLS
    int ISAL_DEF_DIST_SYMBOLS
    int ISAL_DEF_LIT_LEN_SYMBOLS

    # Deflate Implementation Specific Define
    int IGZIP_HIST_SIZE

    # Max repeat length
    int ISAL_LOOK_AHEAD

    # Flush flags
    int NO_FLUSH  # Defaults
    int SYNC_FLUSH
    int FULL_FLUSH
    int FINISH_FLUSH

    # Gzip flags
    int IGZIP_DEFLATE  # Default
    int IGZIP_GZIP
    int IGZIP_GZIP_NO_HDR
    int IGZIP_ZLIB
    int IGZIP_ZLIB_NO_HDR

    # Compression return values
    int COMP_OK 
    int INVALID_FLUSH
    int INVALID_PARAM
    int STATELESS_OVERFLOW
    int ISAL_INVALID_OPERATION
    int ISAL_INVALID_STATE 
    int ISAL_INVALID_LEVEL 
    int ISAL_INVALID_LEVEL_BUF

    cdef enum isal_zstate_state:
        ZSTATE_NEW_HDR  #!< Header to be written
        ZSTATE_HDR,  #!< Header state
        ZSTATE_CREATE_HDR  #!< Header to be created
        ZSTATE_BODY,  #!< Body state
        ZSTATE_FLUSH_READ_BUFFER  #!< Flush buffer
        ZSTATE_FLUSH_ICF_BUFFER
        ZSTATE_TYPE0_HDR  #! Type0 block header to be written
        ZSTATE_TYPE0_BODY  #!< Type0 block body to be written
        ZSTATE_SYNC_FLUSH  #!< Write sync flush block
        ZSTATE_FLUSH_WRITE_BUFFER  #!< Flush bitbuf
        ZSTATE_TRL,  #!< Trailer state
        ZSTATE_END,  #!< End state
        ZSTATE_TMP_NEW_HDR  #!< Temporary Header to be written
        ZSTATE_TMP_HDR,  #!< Temporary Header state
        ZSTATE_TMP_CREATE_HDR  #!< Temporary Header to be created state
        ZSTATE_TMP_BODY  #!< Temporary Body state
        ZSTATE_TMP_FLUSH_READ_BUFFER  #!< Flush buffer
        ZSTATE_TMP_FLUSH_ICF_BUFFER
        ZSTATE_TMP_TYPE0_HDR  #! Temporary Type0 block header to be written
        ZSTATE_TMP_TYPE0_BODY  #!< Temporary Type0 block body to be written
        ZSTATE_TMP_SYNC_FLUSH  #!< Write sync flush block
        ZSTATE_TMP_FLUSH_WRITE_BUFFER  #!< Flush bitbuf
        ZSTATE_TMP_TRL   #!< Temporary Trailer state
        ZSTATE_TMP_END  #!< Temporary End state

    cdef enum isal_block_state:
        ISAL_BLOCK_NEW_HDR,  # Just starting a new block */
        ISAL_BLOCK_HDR,    # In the middle of reading in a block header */
        ISAL_BLOCK_TYPE0,  # Decoding a type 0 block */
        ISAL_BLOCK_CODED,  # Decoding a huffman coded block */
        ISAL_BLOCK_INPUT_DONE,  # Decompression of input is completed */
        ISAL_BLOCK_FINISH,  # Decompression of input is completed and all data has been flushed to output */
        ISAL_GZIP_EXTRA_LEN,
        ISAL_GZIP_EXTRA,
        ISAL_GZIP_NAME,
        ISAL_GZIP_COMMENT,
        ISAL_GZIP_HCRC,
        ISAL_ZLIB_DICT,
        ISAL_CHECKSUM_CHECK,
    
    # Inflate flags
    int ISAL_DEFLATE 
    int ISAL_GZIP 
    int ISAL_GZIP_NO_HDR
    int ISAL_ZLIB
    int ISAL_ZLIB_NO_HDR
    int ISAL_ZLIB_NO_HDR_VER
    int ISAL_GZIP_NO_HDR_VER

    # Inflate return values
    int ISAL_DECOMP_OK  # No errors encountered while decompressing
    int ISAL_END_INPUT  # End of input reached
    int ISAL_OUT_OVERFLOW  # End of output reached
    int ISAL_NAME_OVERFLOW  # End of gzip name buffer reached
    int ISAL_COMMENT_OVERFLOW  # End of gzip name buffer reached
    int ISAL_EXTRA_OVERFLOW  # End of extra buffer reached
    int ISAL_NEED_DICT  # Stream needs a dictionary to continue
    int ISAL_INVALID_BLOCK  # Invalid deflate block found
    int ISAL_INVALID_SYMBOL  # Invalid deflate symbol found
    int ISAL_INVALID_LOOKBACK  # Invalid lookback distance found
    int ISAL_INVALID_WRAPPER  # Invalid gzip/zlib wrapper found
    int ISAL_UNSUPPORTED_METHOD  # Gzip/zlib wrapper specifies unsupported compress method
    int ISAL_INCORRECT_CHECKSUM  # Incorrect checksum found

    # Compression structures
    int ISAL_DEF_MIN_LEVEL
    int ISAL_DEF_MAX_LEVEL

    int ISAL_DEF_LVL0_MIN
    int ISAL_DEF_LVL0_SMALL
    int ISAL_DEF_LVL0_MEDIUM
    int ISAL_DEF_LVL0_LARGE
    int ISAL_DEF_LVL0_EXTRA_LARGE
    int ISAL_DEF_LVL0_DEFAULT

    int ISAL_DEF_LVL1_MIN
    int ISAL_DEF_LVL1_SMALL
    int ISAL_DEF_LVL1_MEDIUM
    int ISAL_DEF_LVL1_LARGE
    int ISAL_DEF_LVL1_EXTRA_LARGE
    int ISAL_DEF_LVL1_DEFAULT

    int ISAL_DEF_LVL2_MIN
    int ISAL_DEF_LVL2_SMALL
    int ISAL_DEF_LVL2_MEDIUM
    int ISAL_DEF_LVL2_LARGE
    int ISAL_DEF_LVL2_EXTRA_LARGE
    int ISAL_DEF_LVL2_DEFAULT

    int ISAL_DEF_LVL3_MIN
    int ISAL_DEF_LVL3_SMALL
    int ISAL_DEF_LVL3_MEDIUM
    int ISAL_DEF_LVL3_LARGE
    int ISAL_DEF_LVL3_EXTRA_LARGE
    int ISAL_DEF_LVL3_DEFAULT
    
    cdef struct BitBuf2:
        unsigned long long m_bits  #!< bits in the bit buffer
        unsigned int m_bits_count;  #!< number of valid bits in the bit buffer
        unsigned char *m_out_buff  #!< current index of buffer to write to
        unsigned char *m_out_end  #!< end of buffer to write to
        unsigned char *m_out_start  #!< start of buffer to write to
    
    cdef struct isal_zstate:
        unsigned int total_in_start #!< Not used, may be replaced with something else
        unsigned int block_next  #!< Start of current deflate block in the input
        unsigned int block_end  #!< End of current deflate block in the input
        unsigned int dist_mask  #!< Distance mask used.
        unsigned int hash_mask
        isal_zstate_state state  #!< Current state in processing the data stream
        BitBuf2 bitbuf
        unsigned int crc  #!< Current checksum without finalize step if any (adler)
        unsigned char has_wrap_hdr  #!< keeps track of wrapper header
        unsigned char has_eob_hdr  #!< keeps track of eob on the last deflate block
        unsigned char has_hist  #!< flag to track if there is match history
        unsigned short has_level_buf_init  #!< flag to track if user supplied memory has been initialized.
        unsigned int count  #!< used for partial header/trailer writes
        unsigned char tmp_out_buff[16]  #! temporary array
        unsigned int tmp_out_start  #!< temporary variable
        unsigned int tmp_out_end  #!< temporary variable
        unsigned int b_bytes_valid  #!< number of valid bytes in buffer
        unsigned int b_bytes_processed  #!< number of bytes processed in buffer

    cdef struct isal_hufftables:
        pass

    cdef struct isal_zstream:
        unsigned char *next_in  #!< Next input byte
        unsigned int avail_in  #!< number of bytes available at next_in
        unsigned int total_in_start  #!< total number of bytes read so far
        unsigned char *next_out  #!< Next output byte
        unsigned int avail_out  #!< number of bytes available at next_out
        unsigned int total_out  #!< total number of bytes written so far
        isal_hufftables *hufftables  #!< Huffman encoding used when compressing
        unsigned int level  #!< Compression level to use
        unsigned int level_buf_size  #!< Size of level_buf
        unsigned char * level_buf  #!< User allocated buffer required for different compression levels
        unsigned short end_of_stream  #!< non-zero if this is the last input buffer
        unsigned short flush  #!< Flush type can be NO_FLUSH, SYNC_FLUSH or FULL_FLUSH
        unsigned short gzip_flag  #!< Indicate if gzip compression is to be performed
        unsigned short hist_bits  #!< Log base 2 of maximum lookback distance, 0 is use default
        isal_zstate internal_state  #!< Internal state for this stream

    # Inflate structures
    cdef struct inflate_huff_code_large:
        pass
    cdef struct inflate_huff_code_small:
        pass

    cdef struct inflate_state:
        unsigned char *next_out  #!< Next output byte
        unsigned int avail_out  #!< number of bytes available at next_out
        unsigned int total_out  #!< total number of bytes written so far
        unsigned char *next_in  #!< Next input byte
        unsigned int avail_in  #!< number of bytes available at next_in
        unsigned long long read_in  #!< Bits buffered to handle unaligned streams
        int read_in_length  #!< Bits in read_in
        inflate_huff_code_large  lit_huff_code  #!< Structure for decoding lit/len symbols
        inflate_huff_code_small  dist_huff_code  #!< Structure for decoding dist symbols
        isal_block_state block_state  #!< Current decompression state
        unsigned int dict_length  #!< Length of dictionary used
        unsigned int bfinal  #!< Flag identifying final block
        unsigned int crc_flag  #!< Flag identifying whether to track of crc
        unsigned int crc  #!< Contains crc or adler32 of output if crc_flag is set
        unsigned int hist_bits  #!< Log base 2 of maximum lookback distance
        # Other members are omitted because they are not in use yet.

    # Compression functions
    # /**
    #  * @brief Initialize compression stream data structure
    #  *
    #  * @param stream Structure holding state information on the compression streams.
    #  * @returns none
    #  */
    cdef void isal_deflate_init(isal_zstream *stream)
    #/**
    #  * @brief Initialize compression stream data structure
    #  *
    #  * @param stream Structure holding state information on the compression streams.
    #  * @returns none
    #  */
    cdef void isal_deflate_stateless_init(isal_zstream *stream)

    #  /**
    #  * @brief Set compression dictionary to use
    #  *
    #  * This function is to be called after isal_deflate_init, or after completing a
    #  * SYNC_FLUSH or FULL_FLUSH and before the next call do isal_deflate. If the
    #  * dictionary is longer than IGZIP_HIST_SIZE, only the last IGZIP_HIST_SIZE
    #  * bytes will be used.
    #  *
    #  * @param stream Structure holding state information on the compression streams.
    #  * @param dict: Array containing dictionary to use.
    #  * @param dict_len: Length of dict.
    #  * @returns COMP_OK,
    #  *          ISAL_INVALID_STATE (dictionary could not be set)
    #  */
    cdef int isal_deflate_set_dict(isal_zstream *stream, 
                                   unsigned char *dict,
                                   unsigned int dict_len )


    #/**
    #  * @brief Fast data (deflate) compression for storage applications.
    #  *
    #  * The call to isal_deflate() will take data from the input buffer (updating
    #  * next_in, avail_in and write a compressed stream to the output buffer
    #  * (updating next_out and avail_out). The function returns when either the input
    #  * buffer is empty or the output buffer is full.
    #  *
    #  * On entry to isal_deflate(), next_in points to an input buffer and avail_in
    #  * indicates the length of that buffer. Similarly next_out points to an empty
    #  * output buffer and avail_out indicates the size of that buffer.
    #  *
    #  * The fields total_in and total_out start at 0 and are updated by
    #  * isal_deflate(). These reflect the total number of bytes read or written so far.
    #  *
    #  * When the last input buffer is passed in, signaled by setting the
    #  * end_of_stream, the routine will complete compression at the end of the input
    #  * buffer, as long as the output buffer is big enough.
    #  *
    #  * The compression level can be set by setting level to any value between
    #  * ISAL_DEF_MIN_LEVEL and ISAL_DEF_MAX_LEVEL. When the compression level is
    #  * ISAL_DEF_MIN_LEVEL, hufftables can be set to a table trained for the the
    #  * specific data type being compressed to achieve better compression. When a
    #  * higher compression level is desired, a larger generic memory buffer needs to
    #  * be supplied by setting level_buf and level_buf_size to represent the chunk of
    #  * memory. For level x, the suggest size for this buffer this buffer is
    #  * ISAL_DEFL_LVLx_DEFAULT. The defines ISAL_DEFL_LVLx_MIN, ISAL_DEFL_LVLx_SMALL,
    #  * ISAL_DEFL_LVLx_MEDIUM, ISAL_DEFL_LVLx_LARGE, and ISAL_DEFL_LVLx_EXTRA_LARGE
    #  * are also provided as other suggested sizes.
    #  *
    #  * The equivalent of the zlib FLUSH_SYNC operation is currently supported.
    #  * Flush types can be NO_FLUSH, SYNC_FLUSH or FULL_FLUSH. Default flush type is
    #  * NO_FLUSH. A SYNC_ OR FULL_ flush will byte align the deflate block by
    #  * appending an empty stored block once all input has been compressed, including
    #  * the buffered input. Checking that the out_buffer is not empty or that
    #  * internal_state.state = ZSTATE_NEW_HDR is sufficient to guarantee all input
    #  * has been flushed. Additionally FULL_FLUSH will ensure look back history does
    #  * not include previous blocks so new blocks are fully independent. Switching
    #  * between flush types is supported.
    #  *
    #  * If a compression dictionary is required, the dictionary can be set calling
    #  * isal_deflate_set_dictionary before calling isal_deflate.
    #  *
    #  * If the gzip_flag is set to IGZIP_GZIP, a generic gzip header and the gzip
    #  * trailer are written around the deflate compressed data. If gzip_flag is set
    #  * to IGZIP_GZIP_NO_HDR, then only the gzip trailer is written. A full-featured
    #  * header is supported by the isal_write_{gzip,zlib}_header() functions.
    #  *
    #  * @param  stream Structure holding state information on the compression streams.
    #  * @return COMP_OK (if everything is ok),
    #  *         INVALID_FLUSH (if an invalid FLUSH is selected),
    #  *         ISAL_INVALID_LEVEL (if an invalid compression level is selected),
    #  *         ISAL_INVALID_LEVEL_BUF (if the level buffer is not large enough).
    #  */
    cdef int isal_deflate(isal_zstream *stream)

    #/**
    #  * @brief Fast data (deflate) stateless compression for storage applications.
    #  *
    #  * Stateless (one shot) compression routine with a similar interface to
    #  * isal_deflate() but operates on entire input buffer at one time. Parameter
    #  * avail_out must be large enough to fit the entire compressed output. Max
    #  * expansion is limited to the input size plus the header size of a stored/raw
    #  * block.
    #  *
    #  * When the compression level is set to 1, unlike in isal_deflate(), level_buf
    #  * may be optionally set depending on what what performance is desired.
    #  *
    #  * For stateless the flush types NO_FLUSH and FULL_FLUSH are supported.
    #  * FULL_FLUSH will byte align the output deflate block so additional blocks can
    #  * be easily appended.
    #  *
    #  * If the gzip_flag is set to IGZIP_GZIP, a generic gzip header and the gzip
    #  * trailer are written around the deflate compressed data. If gzip_flag is set
    #  * to IGZIP_GZIP_NO_HDR, then only the gzip trailer is written.
    #  *
    #  * @param  stream Structure holding state information on the compression streams.
    #  * @return COMP_OK (if everything is ok),
    #  *         INVALID_FLUSH (if an invalid FLUSH is selected),
    #  *         ISAL_INVALID_LEVEL (if an invalid compression level is selected),
    #  *         ISAL_INVALID_LEVEL_BUF (if the level buffer is not large enough),
    #  *         STATELESS_OVERFLOW (if output buffer will not fit output).
    #  */
    cdef int isal_deflate_stateless(isal_zstream *stream)


    ###########################
    # Inflate functions
    ###########################
    # /**
    #  * @brief Initialize decompression state data structure
    #  *
    #  * @param state Structure holding state information on the compression streams.
    #  * @returns none
    #  */
    void isal_inflate_init(inflate_state *state)

    # /**
    #  * @brief Reinitialize decompression state data structure
    #  *
    #  * @param state Structure holding state information on the compression streams.
    #  * @returns none
    #  */
    void isal_inflate_reset(inflate_state *state)

    # /**
    #  * @brief Set decompression dictionary to use
    #  *
    #  * This function is to be called after isal_inflate_init. If the dictionary is
    #  * longer than IGZIP_HIST_SIZE, only the last IGZIP_HIST_SIZE bytes will be
    #  * used.
    #  *
    #  * @param state: Structure holding state information on the decompression stream.
    #  * @param dict: Array containing dictionary to use.
    #  * @param dict_len: Length of dict.
    #  * @returns COMP_OK,
    #  *          ISAL_INVALID_STATE (dictionary could not be set)
    #  */
    int isal_inflate_set_dict(inflate_state *state, unsigned char *dict, unsigned int dict_len)

    # /**
    #  * @brief Fast data (deflate) decompression for storage applications.
    #  *
    #  * On entry to isal_inflate(), next_in points to an input buffer and avail_in
    #  * indicates the length of that buffer. Similarly next_out points to an empty
    #  * output buffer and avail_out indicates the size of that buffer.
    #  *
    #  * The field total_out starts at 0 and is updated by isal_inflate(). This
    #  * reflects the total number of bytes written so far.
    #  *
    #  * The call to isal_inflate() will take data from the input buffer (updating
    #  * next_in, avail_in and write a decompressed stream to the output buffer
    #  * (updating next_out and avail_out). The function returns when the input buffer
    #  * is empty, the output buffer is full, invalid data is found, or in the case of
    #  * zlib formatted data if a dictionary is specified. The current state of the
    #  * decompression on exit can be read from state->block-state.
    #  *
    #  * If the crc_flag is set to ISAL_GZIP_NO_HDR the gzip crc of the output is
    #  * stored in state->crc. Alternatively, if the crc_flag is set to
    #  * ISAL_ZLIB_NO_HDR the adler32 of the output is stored in state->crc (checksum
    #  * may not be updated until decompression is complete). When the crc_flag is set
    #  * to ISAL_GZIP_NO_HDR_VER or ISAL_ZLIB_NO_HDR_VER, the behavior is the same,
    #  * except the checksum is verified with the checksum after immediately following
    #  * the deflate data. If the crc_flag is set to ISAL_GZIP or ISAL_ZLIB, the
    #  * gzip/zlib header is parsed, state->crc is set to the appropriate checksum,
    #  * and the checksum is verified. If the crc_flag is set to ISAL_DEFLATE
    #  * (default), then the data is treated as a raw deflate block.
    #  *
    #  * The element state->hist_bits has values from 0 to 15, where values of 1 to 15
    #  * are the log base 2 size of the matching window and 0 is the default with
    #  * maximum history size.
    #  *
    #  * If a dictionary is required, a call to isal_inflate_set_dict will set the
    #  * dictionary.
    #  *
    #  * @param  state Structure holding state information on the compression streams.
    #  * @return ISAL_DECOMP_OK (if everything is ok),
    #  *         ISAL_INVALID_BLOCK,
    #  *         ISAL_NEED_DICT,
    #  *         ISAL_INVALID_SYMBOL,
    #  *         ISAL_INVALID_LOOKBACK,
    #  *         ISAL_INVALID_WRAPPER,
    #  *         ISAL_UNSUPPORTED_METHOD,
    #  *         ISAL_INCORRECT_CHECKSUM.
    #  */
    int isal_inflate(inflate_state *state)

    ##########################
    # Other functions
    ##########################
    #     /**
    #  * @brief Calculate Adler-32 checksum, runs appropriate version.
    #  *
    #  * This function determines what instruction sets are enabled and selects the
    #  * appropriate version at runtime.
    #  *
    #  * @param init: initial Adler-32 value
    #  * @param buf: buffer to calculate checksum on
    #  * @param len: buffer length in bytes
    #  *
    #  * @returns 32-bit Adler-32 checksum
    #  */
    unsigned int isal_adler32(unsigned int init,
                               const unsigned char *buf, 
                               unsigned long long len)

############################
# python-isal functions
############################
cdef check_isal_deflate_rc(int rc)
cdef check_isal_inflate_rc(int rc)

ctypedef fused stream_or_state:
    isal_zstream
    inflate_state

cdef inline Py_ssize_t py_ssize_t_min(Py_ssize_t a, Py_ssize_t b):
    if a <= b:
        return a
    else:
        return b

cdef Py_ssize_t arrange_output_buffer_with_maximum(stream_or_state *stream,
                                                   unsigned char **buffer,
                                                   Py_ssize_t length,
                                                   Py_ssize_t max_length)


cdef Py_ssize_t arrange_output_buffer(stream_or_state *stream,
                                      unsigned char **buffer,
                                      Py_ssize_t length)

cdef void arrange_input_buffer(stream_or_state *stream, Py_ssize_t *remains)

cdef:
    int MEM_LEVEL_DEFAULT_I
    int MEM_LEVEL_MIN_I
    int MEM_LEVEL_SMALL_I
    int MEM_LEVEL_MEDIUM_I
    int MEM_LEVEL_LARGE_I
    int MEM_LEVEL_EXTRA_LARGE_I
    int ISAL_DEFAULT_COMPRESSION_I

cdef int mem_level_to_bufsize(int compression_level, int mem_level, unsigned int *bufsize)

cdef _compress(data,
             int level,
             int flag,
             int mem_level,
             int hist_bits,
            )

cdef _decompress(data,
                 int flag,
                 int hist_bits,
                 Py_ssize_t bufsize)

cdef bytes view_bitbuffer(inflate_state * stream)


# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

ISAL_BEST_SPEED: int
ISAL_BEST_COMPRESSION: int
ISAL_DEFAULT_COMPRESSION: int
DEF_BUF_SIZE: int
MAX_HIST_BITS: int
ISAL_NO_FLUSH: int
ISAL_SYNC_FLUSH: int
ISAL_FULL_FLUSH: int
COMP_DEFLATE: int
COMP_GZIP: int
COMP_GZIP_NO_HDR: int
COMP_ZLIB: int
COMP_ZLIB_NO_HDR: int
DECOMP_DEFLATE: int
DECOMP_ZLIB: int
DECOMP_GZIP: int
DECOMP_GZIP_NO_HDR: int
DECOMP_ZLIB_NO_HDR: int
DECOMP_ZLIB_NO_HDR_VER: int
DECOMP_GZIP_NO_HDR_VER: int
MEM_LEVEL_DEFAULT: int
MEM_LEVEL_MIN: int
MEM_LEVEL_SMALL: int
MEM_LEVEL_MEDIUM: int
MEM_LEVEL_LARGE: int
MEM_LEVEL_EXTRA_LARGE: int
IsalError: OSError

def compress(data, level: int = ISAL_DEFAULT_COMPRESSION,
             flag: int = COMP_DEFLATE,
             mem_level: int = MEM_LEVEL_DEFAULT,
             hist_bits: int = MAX_HIST_BITS) -> bytes: ...
def decompress(data, flag: int = DECOMP_DEFLATE,
               hist_bits: int = MAX_HIST_BITS,
               bufsize: int = DEF_BUF_SIZE) -> bytes: ...

class IgzipDecompressor:
    unused_data: bytes
    needs_input: bool
    eof: bool

    def decompress(self, data, max_length = -1) -> bytes: ...



# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cython: language_level=3
# cython: binding=True

"""
Pythonic interface to ISA-L's igzip_lib.

This module comes with the following constants:

============================== ================================================
``ISAL_BEST_SPEED``            The lowest compression level (0)
``ISAL_BEST_COMPRESSION``      The highest compression level (3)
 ``ISAL_DEFAULT_COMPRESSION``  The compromise compression level (2)
``DEF_BUF_SIZE``               Default size for the starting buffer (16K)
``MAX_HIST_BITS``              Maximum window size bits (15).
``COMP_DEFLATE``               Flag to compress to a raw deflate block
``COMP_GZIP``                  Flag to compress a gzip block, consisting of a
                               gzip header, raw deflate block and a gzip
                               trailer.
``COMP_GZIP_NO_HDR``           Flag to compress a gzip block without a header.
``COMP_ZLIB``                  Flag to compress a zlib block, consisting of a
                               zlib header, a raw deflate block and a zlib
                               trailer.
``COMP_ZLIB_NO_HDR``           Flag to compress a zlib block without a header.
``DECOMP_DEFLATE``             Flag to decompress a raw deflate block.
``DECOMP_GZIP``                Flag to decompress a gzip block including header
                               and verify the checksums in the trailer.
``DECOMP_GZIP_NO_HDR``         Flag to decompress a gzip block without a header
                               and verify the checksums in the trailer.
``DECOMP_GZIP_NO_HDR_VER``     Flag to decompress a gzip block without a header
                               and without verifying the checksums in the
                               trailer.
``DECOMP_ZLIB``                Flag to decompress a zlib block including header
                               and verify the checksums in the trailer.
``DECOMP_ZLIB_NO_HDR``         Flag to decompress a zlib block without a header
                               and verify the checksums in the trailer.
``DECOMP_ZLIB_NO_HDR_VER``     Flag to decompress a zlib block without a header
                               and without verifying the checksums in the
                               trailer.
``MEM_LEVEL_DEFAULT``          The default memory level for the internal level
                               buffer. (Equivalent to
                               MEM_LEVEL_LARGE.)
``MEM_LEVEL_MIN``              The minimum memory level.
``MEM_LEVEL_SMALL``
``MEM_LEVEL_MEDIUM``
``MEM_LEVEL_LARGE``
``MEM_LEVEL_EXTRA_LARGE``      The largest memory level.
============================== ================================================
"""

from libc.stdint cimport UINT64_MAX, UINT32_MAX
from libc.string cimport memmove, memcpy
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython.buffer cimport PyBUF_C_CONTIGUOUS, PyObject_GetBuffer, PyBuffer_Release
from cpython.bytes cimport PyBytes_FromStringAndSize

cdef extern from "<Python.h>":
    const Py_ssize_t PY_SSIZE_T_MAX

ISAL_BEST_SPEED = ISAL_DEF_MIN_LEVEL
ISAL_BEST_COMPRESSION = ISAL_DEF_MAX_LEVEL
cdef int ISAL_DEFAULT_COMPRESSION_I = 2
ISAL_DEFAULT_COMPRESSION = ISAL_DEFAULT_COMPRESSION_I

DEF DEF_BUF_SIZE_I = 16 * 1024
DEF_BUF_SIZE = DEF_BUF_SIZE_I
MAX_HIST_BITS = ISAL_DEF_MAX_HIST_BITS

ISAL_NO_FLUSH = NO_FLUSH
ISAL_SYNC_FLUSH = SYNC_FLUSH
ISAL_FULL_FLUSH = FULL_FLUSH

COMP_DEFLATE = IGZIP_DEFLATE
COMP_GZIP = IGZIP_GZIP
COMP_GZIP_NO_HDR = IGZIP_GZIP_NO_HDR
COMP_ZLIB = IGZIP_ZLIB
COMP_ZLIB_NO_HDR = IGZIP_ZLIB_NO_HDR

DECOMP_DEFLATE = ISAL_DEFLATE
DECOMP_GZIP = ISAL_GZIP
DECOMP_GZIP_NO_HDR = ISAL_GZIP_NO_HDR
DECOMP_ZLIB = ISAL_ZLIB
DECOMP_ZLIB_NO_HDR = ISAL_ZLIB_NO_HDR
DECOMP_ZLIB_NO_HDR_VER = ISAL_ZLIB_NO_HDR_VER
DECOMP_GZIP_NO_HDR_VER = ISAL_GZIP_NO_HDR_VER

cdef int MEM_LEVEL_DEFAULT_I = 0
cdef int MEM_LEVEL_MIN_I = 1
cdef int MEM_LEVEL_SMALL_I = 2
cdef int MEM_LEVEL_MEDIUM_I = 3
cdef int MEM_LEVEL_LARGE_I = 4
cdef int MEM_LEVEL_EXTRA_LARGE_I = 5
MEM_LEVEL_DEFAULT = MEM_LEVEL_DEFAULT_I
MEM_LEVEL_MIN = MEM_LEVEL_MIN_I
MEM_LEVEL_SMALL = MEM_LEVEL_SMALL_I
MEM_LEVEL_MEDIUM = MEM_LEVEL_MEDIUM_I
MEM_LEVEL_LARGE = MEM_LEVEL_LARGE_I
MEM_LEVEL_EXTRA_LARGE = MEM_LEVEL_EXTRA_LARGE_I

class IsalError(OSError):
    """Exception raised on compression and decompression errors."""
    pass



cdef Py_ssize_t arrange_output_buffer_with_maximum(stream_or_state *stream,
                                                   unsigned char **buffer,
                                                   Py_ssize_t length,
                                                   Py_ssize_t max_length):
    # The zlibmodule.c function builds a PyBytes object. A unsigned char *
    # is build here because building raw PyObject * stuff in cython is somewhat
    # harder due to cython's interference. FIXME
    cdef Py_ssize_t occupied
    cdef Py_ssize_t new_length
    cdef unsigned char * new_buffer
    if buffer[0] == NULL:
        buffer[0] = <unsigned char*>PyMem_Malloc(length * sizeof(char))
        if buffer[0] == NULL:
            return -1
        occupied = 0
    else:
        occupied = stream.next_out - buffer[0]
        if length == occupied:
            if length == max_length:
                return -2
            if length <= max_length >> 1:
                new_length = length << 1
            else:
                new_length = max_length
            new_buffer = <unsigned char *>PyMem_Realloc(buffer[0], new_length)
            if new_buffer == NULL:
                return -1
            buffer[0] = new_buffer
            length = new_length
    stream.avail_out = <unsigned int>py_ssize_t_min(length - occupied, UINT32_MAX)
    stream.next_out = buffer[0] + occupied
    return length

cdef Py_ssize_t arrange_output_buffer(stream_or_state *stream,
                                      unsigned char **buffer,
                                      Py_ssize_t length):
    cdef Py_ssize_t ret
    ret = arrange_output_buffer_with_maximum(stream, buffer, length, PY_SSIZE_T_MAX)
    if ret == -2:  # Maximum reached.
        return -1
    return ret

cdef void arrange_input_buffer(stream_or_state *stream, Py_ssize_t *remains):
    stream.avail_in = <unsigned int>py_ssize_t_min(remains[0], UINT32_MAX)
    remains[0] -= stream.avail_in

def compress(data,
             int level=ISAL_DEFAULT_COMPRESSION_I,
             int flag = IGZIP_DEFLATE,
             int mem_level = MEM_LEVEL_DEFAULT_I,
             int hist_bits = ISAL_DEF_MAX_HIST_BITS,
            ):
    """
    Compresses the bytes in *data*. Returns a bytes object with the
    compressed data.

    :param level: the compression level from 0 to 3. 0 is the lowest
                  compression (NOT no compression as in stdlib zlib!) and the
                  fastest. 3 is the best compression and the slowest. Default
                  is a compromise at level 2.
    :param flag:  Controls the header and trailer. Can be any of: COMP_DEFLATE
                  (default), COMP_GZIP, COMP_GZIP_NO_HDR, COMP_ZLIB, 
                  COMP_ZLIB_NO_HDR.
    :param mem_level: Set the memory level for the memory buffer. Larger
                      buffers improve performance. Can be any of: 
                      MEM_LEVEL_DEFAULT (default, 
                      equivalent to MEM_LEVEL_LARGE), MEM_LEVEL_MIN, 
                      MEM_LEVEL_SMALL, MEM_LEVEL_MEDIUM, MEM_LEVEL_LARGE,
                      MEM_LEVEL_EXTRA_LARGE.
    :param hist_bits: Sets the size of the view window. The size equals 
                      2^hist_bits. Similar to zlib wbits value, except that 
                      hist_bits is not used to set the compression flag.
                      This is best left at the default (15, maximum).
    """
    return _compress(data, level, flag, mem_level, hist_bits)


cdef _compress(data,
             int level,
             int flag,
             int mem_level,
             int hist_bits,
            ):
    # Initialise stream
    cdef isal_zstream stream
    cdef unsigned int level_buf_size
    mem_level_to_bufsize(level, MEM_LEVEL_DEFAULT_I, &level_buf_size)
    cdef unsigned char* level_buf = <unsigned char*> PyMem_Malloc(level_buf_size * sizeof(char))
    isal_deflate_init(&stream)
    stream.level = level
    stream.level_buf = level_buf
    stream.level_buf_size = level_buf_size
    stream.hist_bits = hist_bits
    stream.gzip_flag = flag

    # Initialise output buffer
    cdef Py_ssize_t bufsize = DEF_BUF_SIZE_I
    cdef unsigned char * obuf = NULL
    
    # initialise input
    cdef Py_buffer buffer_data
    cdef Py_buffer* buffer = &buffer_data
    # Cython makes sure error is handled when acquiring buffer fails.
    PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
    cdef Py_ssize_t ibuflen = buffer.len
    stream.next_in = <unsigned char*>buffer.buf

    # initialise helper variables
    cdef int err

    try:
        while True:
            arrange_input_buffer(&stream, &ibuflen)
            if ibuflen == 0:
                stream.flush = FULL_FLUSH
                stream.end_of_stream = 1
            else:
                stream.flush = NO_FLUSH

            while True:
                bufsize = arrange_output_buffer(&stream, &obuf, bufsize)
                if bufsize == -1:
                    raise MemoryError("Unsufficient memory for buffer allocation")
                err = isal_deflate(&stream)
                if err != COMP_OK:
                    check_isal_deflate_rc(err)
                if stream.avail_out != 0:
                    break
            if stream.avail_in != 0:
                raise AssertionError("Input stream should be empty")
            if stream.internal_state.state == ZSTATE_END:
                break
        return PyBytes_FromStringAndSize(<char*>obuf, stream.next_out - obuf)
    finally:
        PyBuffer_Release(buffer)
        PyMem_Free(level_buf)
        PyMem_Free(obuf)


def decompress(data,
                 int flag = ISAL_DEFLATE,
                 int hist_bits=ISAL_DEF_MAX_HIST_BITS,
                 Py_ssize_t bufsize=DEF_BUF_SIZE):
    """
    Deompresses the bytes in *data*. Returns a bytes object with the
    decompressed data.

    :param flag: Whether the compressed block contains headers and/or trailers
                 and of which type. Can be any of: DECOMP_DEFLATE (default), 
                 DECOMP_GZIP, DECOMP_GZIP_NO_HDR, DECOMP_GZIP_NO_HDR_VER,
                 DECOMP_ZLIB, DECOMP_ZLIB_NO_HDR, DECOMP_ZLIB_NO_HDR_VER.
    :param hist_bits: Sets the size of the view window. The size equals 
                      2^hist_bits. Similar to zlib wbits value, except that 
                      hist_bits is not used to set the compression flag.
                      This is best left at the default (15, maximum).
    :param bufsize: The initial size of the output buffer. The output buffer
                    is dynamically resized according to the need. The default
                    size is 16K. If a larger output is expected, using a 
                    larger buffer will improve performance by negating the 
                    costs associated with the dynamic resizing.
    """
    return _decompress(data, flag, hist_bits, bufsize)


cdef _decompress(data,
                 int flag,
                 int hist_bits,
                 Py_ssize_t bufsize):
    if bufsize < 0:
        raise ValueError("bufsize must be non-negative")
   
    cdef inflate_state stream
    isal_inflate_init(&stream)
    stream.hist_bits = hist_bits
    stream.crc_flag = flag

    # initialise input
    cdef Py_buffer buffer_data
    cdef Py_buffer* buffer = &buffer_data
    # Cython makes sure error is handled when acquiring buffer fails.
    PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
    cdef Py_ssize_t ibuflen = buffer.len
    stream.next_in =  <unsigned char*>buffer.buf

    # Initialise output buffer
    cdef unsigned char * obuf = NULL
    cdef int err

    try:
        while True:
            arrange_input_buffer(&stream, &ibuflen)

            while True:
                bufsize = arrange_output_buffer(&stream, &obuf, bufsize)
                if bufsize == -1:
                    raise MemoryError("Unsufficient memory for buffer allocation")
                err = isal_inflate(&stream)
                if err != ISAL_DECOMP_OK:
                    check_isal_inflate_rc(err)
                if stream.avail_out != 0:
                    break
            if ibuflen == 0 or stream.block_state == ISAL_BLOCK_FINISH:
                break
        if stream.block_state != ISAL_BLOCK_FINISH:
            raise IsalError("incomplete or truncated stream")
        return PyBytes_FromStringAndSize(<char*>obuf, stream.next_out - obuf)
    finally:
        PyBuffer_Release(buffer)
        PyMem_Free(obuf)


cdef bytes view_bitbuffer(inflate_state * stream):

        cdef int bits_in_buffer = stream.read_in_length
        cdef int read_in_length = bits_in_buffer // 8
        if read_in_length == 0:
            return b""
        cdef int remainder = bits_in_buffer % 8
        read_in = stream.read_in
        # The bytes are added by bitshifting, so in reverse order. Reading the
        # 64-bit integer into 8 bytes little-endian provides the characters in
        # the correct order.
        return (read_in >> remainder).to_bytes(8, "little")[:read_in_length]


cdef class IgzipDecompressor:
    """Decompress object for handling streaming decompression."""
    cdef public bytes unused_data
    cdef public bint eof
    cdef public bint needs_input
    cdef inflate_state stream
    cdef unsigned char * input_buffer
    cdef size_t input_buffer_size
    cdef Py_ssize_t avail_in_real

    def __dealloc__(self):
        if self.input_buffer != NULL:
            PyMem_Free(self.input_buffer)

    def __cinit__(self,
                  flag=ISAL_DEFLATE,
                  hist_bits=ISAL_DEF_MAX_HIST_BITS,
                  zdict = None):
        isal_inflate_init(&self.stream)

        self.stream.hist_bits = hist_bits
        self.stream.crc_flag = flag
        cdef Py_ssize_t zdict_length
        if zdict:
            zdict_length = len(zdict)
            if zdict_length > UINT32_MAX:
                raise OverflowError("zdict length does not fit in an unsigned int")
            err = isal_inflate_set_dict(&self.stream, zdict, zdict_length)
            if err != COMP_OK:
                check_isal_deflate_rc(err)
        self.unused_data = b""
        self.eof = False
        self.input_buffer = NULL
        self.input_buffer_size = 0
        self.avail_in_real = 0
        self.needs_input = True
        
    def _view_bitbuffer(self):
        """Shows the 64-bitbuffer of the internal inflate_state. It contains
        a maximum of 8 bytes. This data is already read-in so is not part
        of the unconsumed tail."""
        return view_bitbuffer(&self.stream)

    cdef decompress_buf(self, Py_ssize_t max_length, unsigned char ** obuf):
        obuf[0] = NULL
        cdef Py_ssize_t obuflen = DEF_BUF_SIZE_I
        cdef int err
        if obuflen > max_length:
            obuflen = max_length
        while True:
            obuflen = arrange_output_buffer_with_maximum(&self.stream, obuf, obuflen, max_length)
            if obuflen == -1:
                raise MemoryError("Unsufficient memory for buffer allocation")
            elif obuflen == -2:
                break
            arrange_input_buffer(&self.stream, &self.avail_in_real)
            err = isal_inflate(&self.stream)
            self.avail_in_real += self.stream.avail_in
            if err != ISAL_DECOMP_OK:
                check_isal_inflate_rc(err)
            if self.stream.block_state == ISAL_BLOCK_FINISH:
                self.eof = 1
                break
            elif self.avail_in_real == 0:
                break
        return

    def decompress(self, data, Py_ssize_t max_length = -1):
        """
        Decompress data, returning a bytes object containing the uncompressed
        data corresponding to at least part of the data in string.

        :param data: Binary data (bytes, bytearray, memoryview).
        :param max_length: if non-zero then the return value will be no longer
                           than max_length.
        """
        if self.eof:
            raise EOFError("End of stream already reached")
        cdef bint input_buffer_in_use
        
        cdef Py_ssize_t hard_limit
        if max_length < 0:
            hard_limit = PY_SSIZE_T_MAX
        else:
            hard_limit = max_length

        cdef unsigned int avail_now
        cdef unsigned int avail_total
        # Cython makes sure error is handled when acquiring buffer fails.
        cdef Py_buffer buffer_data
        cdef Py_buffer* buffer = &buffer_data
        PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
        cdef Py_ssize_t ibuflen = buffer.len
        cdef unsigned char * data_ptr = <unsigned char*>buffer.buf


        cdef bint max_length_reached = False
        cdef unsigned char * tmp
        cdef size_t offset
        # Initialise output buffer
        cdef unsigned char *obuf = NULL

        try:
            if self.stream.next_in != NULL:
                avail_now = (self.input_buffer + self.input_buffer_size) - \
                            (self.stream.next_in + self.avail_in_real)
                avail_total = self.input_buffer_size - self.avail_in_real
                if avail_total < ibuflen:
                    offset = self.stream.next_in - self.input_buffer
                    new_size = self.input_buffer_size + ibuflen - avail_now
                    tmp = <unsigned char*>PyMem_Realloc(self.input_buffer, new_size)
                    if tmp == NULL:
                        raise MemoryError()
                    self.input_buffer = tmp
                    self.input_buffer_size = new_size
                    self.stream.next_in = self.input_buffer + offset
                elif avail_now < ibuflen:
                    memmove(self.input_buffer, self.stream.next_in,
                            self.avail_in_real)
                    self.stream.next_in = self.input_buffer
                memcpy(<void *>(self.stream.next_in + self.avail_in_real), data_ptr, buffer.len)
                self.avail_in_real += ibuflen
                input_buffer_in_use = 1
            else:
                self.stream.next_in = data_ptr
                self.avail_in_real = ibuflen
                input_buffer_in_use = 0

            self.decompress_buf(hard_limit, &obuf)
            if obuf == NULL:
                self.stream.next_in = NULL
                return b""
            if self.eof:
                self.needs_input = False
                new_data = PyBytes_FromStringAndSize(<char *>self.stream.next_in, self.avail_in_real)
                self.unused_data = self._view_bitbuffer() + new_data
            elif self.avail_in_real == 0:
                self.stream.next_in = NULL
                self.needs_input = True
            else:
                self.needs_input = False
                if not input_buffer_in_use:
                    # Discard buffer if to small.
                    # Resizing may needlessly copy the current contents.
                    if self.input_buffer != NULL and self.input_buffer_size < self.avail_in_real:
                        PyMem_Free(self.input_buffer)
                        self.input_buffer = NULL

                    # Allocate of necessary
                    if self.input_buffer == NULL:
                        self.input_buffer = <unsigned char *>PyMem_Malloc(self.avail_in_real)
                        if self.input_buffer == NULL:
                            raise MemoryError()
                        self.input_buffer_size = self.avail_in_real

                    # Copy tail
                    memcpy(self.input_buffer, self.stream.next_in, self.avail_in_real)
                    self.stream.next_in = self.input_buffer
            return PyBytes_FromStringAndSize(<char*>obuf, self.stream.next_out - obuf)
        except:
            self.stream.next_in = NULL
            raise
        finally:
            PyBuffer_Release(buffer)
            PyMem_Free(obuf)


cdef int mem_level_to_bufsize(int compression_level, int mem_level, unsigned int *bufsize):
    """
    Convert zlib memory levels to isal equivalents
    """
    if not (0 <= mem_level <= MEM_LEVEL_EXTRA_LARGE_I):
        bufsize[0] = 0
        return -1
    if not (ISAL_DEF_MIN_LEVEL <= compression_level <= ISAL_DEF_MAX_LEVEL):
        bufsize[0] = 0
        return -1

    if mem_level == MEM_LEVEL_DEFAULT_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_DEFAULT
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_DEFAULT
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_DEFAULT
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_DEFAULT
    elif mem_level == MEM_LEVEL_MIN_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_MIN
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_MIN
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_MIN
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_MIN
    elif mem_level == MEM_LEVEL_SMALL_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_SMALL
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_SMALL
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_SMALL
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_SMALL
    elif mem_level == MEM_LEVEL_MEDIUM_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_MEDIUM
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_MEDIUM
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_MEDIUM
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_MEDIUM
    elif mem_level == MEM_LEVEL_LARGE_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_LARGE
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_LARGE
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_LARGE
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_LARGE
    elif mem_level == MEM_LEVEL_EXTRA_LARGE_I:
        if compression_level == 0:
            bufsize[0] = ISAL_DEF_LVL0_EXTRA_LARGE
        elif compression_level == 1:
            bufsize[0] = ISAL_DEF_LVL1_EXTRA_LARGE
        elif compression_level == 2:
            bufsize[0] = ISAL_DEF_LVL2_EXTRA_LARGE
        elif compression_level == 3:
            bufsize[0] = ISAL_DEF_LVL3_EXTRA_LARGE
    return 0

cdef check_isal_deflate_rc(int rc):
    if rc == COMP_OK:
        return
    elif rc == INVALID_FLUSH:
        raise IsalError("Invalid flush type")
    elif rc == INVALID_PARAM:
        raise IsalError("Invalid parameter")
    elif rc == STATELESS_OVERFLOW:
        raise IsalError("Not enough room in output buffer")
    elif rc == ISAL_INVALID_OPERATION:
        raise IsalError("Invalid operation")
    elif rc == ISAL_INVALID_STATE:
        raise IsalError("Invalid state")
    elif rc == ISAL_INVALID_LEVEL:
        raise IsalError("Invalid compression level.")
    elif rc == ISAL_INVALID_LEVEL_BUF:
        raise IsalError("Level buffer too small.")
    else:
        raise IsalError("Unknown Error")

cdef check_isal_inflate_rc(int rc):
    if rc >= ISAL_DECOMP_OK:
        return
    if rc == ISAL_END_INPUT:
        raise IsalError("End of input reached")
    if rc == ISAL_OUT_OVERFLOW:
        raise IsalError("End of output reached")
    if rc == ISAL_NAME_OVERFLOW:
        raise IsalError("End of gzip name buffer reached")
    if rc == ISAL_COMMENT_OVERFLOW:
        raise IsalError("End of gzip name buffer reached")
    if rc == ISAL_EXTRA_OVERFLOW:
        raise IsalError("End of extra buffer reached")
    if rc == ISAL_NEED_DICT:
        raise IsalError("Dictionary needed to continue")
    if rc == ISAL_INVALID_BLOCK:
        raise IsalError("Invalid deflate block found")
    if rc == ISAL_INVALID_SYMBOL:
        raise IsalError("Invalid deflate symbol found")
    if rc == ISAL_INVALID_LOOKBACK:
        raise IsalError("Invalid lookback distance found")
    if rc == ISAL_INVALID_WRAPPER:
        raise IsalError("Invalid gzip/zlib wrapper found")
    if rc == ISAL_UNSUPPORTED_METHOD:
        raise IsalError("Gzip/zlib wrapper specifies unsupported compress method")
    if rc == ISAL_INCORRECT_CHECKSUM:
        raise IsalError("Incorrect checksum found")


# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Similar to the stdlib gzip module. But using the Intel Storage Accelaration
Library to speed up its methods."""

import argparse
import gzip
import io
import os
import struct
import sys
import time
from typing import List, Optional, SupportsInt
import _compression  # noqa: I201  # Not third-party

from . import igzip_lib, isal_zlib

__all__ = ["IGzipFile", "open", "compress", "decompress", "BadGzipFile",
           "READ_BUFFER_SIZE"]

_COMPRESS_LEVEL_FAST = isal_zlib.ISAL_BEST_SPEED
_COMPRESS_LEVEL_TRADEOFF = isal_zlib.ISAL_DEFAULT_COMPRESSION
_COMPRESS_LEVEL_BEST = isal_zlib.ISAL_BEST_COMPRESSION

#: The amount of data that is read in at once when decompressing a file.
#: Increasing this value may increase performance.
READ_BUFFER_SIZE = io.DEFAULT_BUFFER_SIZE

FTEXT, FHCRC, FEXTRA, FNAME, FCOMMENT = 1, 2, 4, 8, 16

try:
    BadGzipFile = gzip.BadGzipFile  # type: ignore
except AttributeError:  # Versions lower than 3.8 do not have BadGzipFile
    BadGzipFile = OSError


# The open method was copied from the CPython source with minor adjustments.
def open(filename, mode="rb", compresslevel=_COMPRESS_LEVEL_TRADEOFF,
         encoding=None, errors=None, newline=None):
    """Open a gzip-compressed file in binary or text mode. This uses the isa-l
    library for optimized speed.

    The filename argument can be an actual filename (a str or bytes object), or
    an existing file object to read from or write to.

    The mode argument can be "r", "rb", "w", "wb", "x", "xb", "a" or "ab" for
    binary mode, or "rt", "wt", "xt" or "at" for text mode. The default mode is
    "rb", and the default compresslevel is 2.

    For binary mode, this function is equivalent to the GzipFile constructor:
    GzipFile(filename, mode, compresslevel). In this case, the encoding, errors
    and newline arguments must not be provided.

    For text mode, a GzipFile object is created, and wrapped in an
    io.TextIOWrapper instance with the specified encoding, error handling
    behavior, and line ending(s).

    """
    if "t" in mode:
        if "b" in mode:
            raise ValueError("Invalid mode: %r" % (mode,))
    else:
        if encoding is not None:
            raise ValueError(
                "Argument 'encoding' not supported in binary mode")
        if errors is not None:
            raise ValueError("Argument 'errors' not supported in binary mode")
        if newline is not None:
            raise ValueError("Argument 'newline' not supported in binary mode")

    gz_mode = mode.replace("t", "")
    # __fspath__ method is os.PathLike
    if isinstance(filename, (str, bytes)) or hasattr(filename, "__fspath__"):
        binary_file = IGzipFile(filename, gz_mode, compresslevel)
    elif hasattr(filename, "read") or hasattr(filename, "write"):
        binary_file = IGzipFile(None, gz_mode, compresslevel, filename)
    else:
        raise TypeError("filename must be a str or bytes object, or a file")

    if "t" in mode:
        return io.TextIOWrapper(binary_file, encoding, errors, newline)
    else:
        return binary_file


class IGzipFile(gzip.GzipFile):
    """The IGzipFile class simulates most of the methods of a file object with
    the exception of the truncate() method.

    This class only supports opening files in binary mode. If you need to open
    a compressed file in text mode, use the gzip.open() function.
    """
    def __init__(self, filename=None, mode=None,
                 compresslevel=isal_zlib.ISAL_DEFAULT_COMPRESSION,
                 fileobj=None, mtime=None):
        """Constructor for the IGzipFile class.

        At least one of fileobj and filename must be given a
        non-trivial value.

        The new class instance is based on fileobj, which can be a regular
        file, an io.BytesIO object, or any other object which simulates a file.
        It defaults to None, in which case filename is opened to provide
        a file object.

        When fileobj is not None, the filename argument is only used to be
        included in the gzip file header, which may include the original
        filename of the uncompressed file.  It defaults to the filename of
        fileobj, if discernible; otherwise, it defaults to the empty string,
        and in this case the original filename is not included in the header.

        The mode argument can be any of 'r', 'rb', 'a', 'ab', 'w', 'wb', 'x',
        or 'xb' depending on whether the file will be read or written.
        The default is the mode of fileobj if discernible; otherwise, the
        default is 'rb'. A mode of 'r' is equivalent to one of 'rb', and
        similarly for 'w' and 'wb', 'a' and 'ab', and 'x' and 'xb'.

        The compresslevel argument is an integer from 0 to 3 controlling the
        level of compression; 0 is fastest and produces the least compression,
        and 3 is slowest and produces the most compression. Unlike
        gzip.GzipFile 0 is NOT no compression. The default is 2.

        The mtime argument is an optional numeric timestamp to be written
        to the last modification time field in the stream when compressing.
        If omitted or None, the current time is used.
        """
        if not (isal_zlib.ISAL_BEST_SPEED <= compresslevel
                <= isal_zlib.ISAL_BEST_COMPRESSION):
            raise ValueError(
                "Compression level should be between {0} and {1}.".format(
                    isal_zlib.ISAL_BEST_SPEED, isal_zlib.ISAL_BEST_COMPRESSION
                ))
        super().__init__(filename, mode, compresslevel, fileobj, mtime)
        if self.mode == gzip.WRITE:
            self.compress = isal_zlib.compressobj(compresslevel,
                                                  isal_zlib.DEFLATED,
                                                  -isal_zlib.MAX_WBITS,
                                                  isal_zlib.DEF_MEM_LEVEL,
                                                  0)
        if self.mode == gzip.READ:
            raw = _IGzipReader(self.fileobj)
            self._buffer = io.BufferedReader(raw)

    def __repr__(self):
        s = repr(self.fileobj)
        return '<igzip ' + s[1:-1] + ' ' + hex(id(self)) + '>'

    def _write_gzip_header(self, compresslevel=_COMPRESS_LEVEL_TRADEOFF):
        # Python 3.9 added a `compresslevel` parameter to write gzip header.
        # This only determines the value of one extra flag. Because this change
        # was backported to 3.7 and 3.8 in later point versions, the attributes
        # of the function should be checked before trying to use the
        # compresslevel parameter.
        # The gzip header has an extra flag that can be set depending on the
        # compression level used. This should be set when either the fastest or
        # best method is used. ISAL level 0 is larger than gzip level 1 and
        # much faster, so setting the flag for fastest level is appropriate.
        # ISAL level 1,2 and 3 (best)are similar in size and fall around the
        # gzip level 3 size. So setting no extra flag
        # (by using COMPRESS_LEVEL_TRADEOFF) is appropriate here.
        if ("compresslevel" in super()._write_gzip_header.__code__.co_varnames
            and hasattr(gzip, "_COMPRESS_LEVEL_FAST")
                and hasattr(gzip, "_COMPRESS_LEVEL_TRADEOFF")):
            if compresslevel == _COMPRESS_LEVEL_FAST:
                super()._write_gzip_header(gzip._COMPRESS_LEVEL_FAST)
            else:
                super()._write_gzip_header(gzip._COMPRESS_LEVEL_TRADEOFF)
        else:
            super()._write_gzip_header()

    def write(self, data):
        self._check_not_closed()
        if self.mode != gzip.WRITE:
            import errno
            raise OSError(errno.EBADF, "write() on read-only IGzipFile object")

        if self.fileobj is None:
            raise ValueError("write() on closed IGzipFile object")

        if isinstance(data, bytes):
            length = len(data)
        else:
            # accept any data that supports the buffer protocol
            data = memoryview(data)
            length = data.nbytes

        if length > 0:
            self.fileobj.write(self.compress.compress(data))
            self.size += length
            self.crc = isal_zlib.crc32(data, self.crc)
            self.offset += length
        return length


class _PaddedFile(gzip._PaddedFile):
    # Overwrite _PaddedFile from gzip as its prepend method assumes that
    # the prepended data is always read from its _buffer. Unfortunately in
    # isal_zlib.decompressobj there is a bitbuffer as well which may be added.
    # So an extra check is added to prepend to ensure no extra data in front
    # of the buffer was present. (Negative self._read).
    def prepend(self, prepend=b''):
        if self._read is not None:
            # Assume data was read since the last prepend() call
            self._read -= len(prepend)
            if self._read >= 0:
                return
            # If self._read is negative the data was read further back and
            # the buffer needs to be reset.
        self._buffer = prepend
        self._length = len(self._buffer)
        self._read = 0


class _IGzipReader(gzip._GzipReader):
    def __init__(self, fp):
        # Call the init method of gzip._GzipReader's parent here.
        # It is not very invasive and allows us to override _PaddedFile
        _compression.DecompressReader.__init__(
            self, _PaddedFile(fp), igzip_lib.IgzipDecompressor,
            hist_bits=igzip_lib.MAX_HIST_BITS, flag=igzip_lib.DECOMP_DEFLATE)
        # Set flag indicating start of a new member
        self._new_member = True
        self._last_mtime = None

    def _add_read_data(self, data):
        # Use faster isal crc32 calculation and update the stream size in place
        # compared to CPython gzip
        self._crc = isal_zlib.crc32(data, self._crc)
        self._stream_size += len(data)

    def read(self, size=-1):
        if size < 0:
            return self.readall()
        # size=0 is special because decompress(max_length=0) is not supported
        if not size:
            return b""

        # For certain input data, a single
        # call to decompress() may not return
        # any data. In this case, retry until we get some data or reach EOF.
        while True:
            if self._decompressor.eof:
                # Ending case: we've come to the end of a member in the file,
                # so finish up this member, and read a new gzip header.
                # Check the CRC and file size, and set the flag so we read
                # a new member
                self._read_eof()
                self._new_member = True
                self._decompressor = self._decomp_factory(
                    **self._decomp_args)

            if self._new_member:
                # If the _new_member flag is set, we have to
                # jump to the next member, if there is one.
                self._init_read()
                if not self._read_gzip_header():
                    self._size = self._pos
                    return b""
                self._new_member = False

            # Read a chunk of data from the file
            if self._decompressor.needs_input:
                buf = self._fp.read(READ_BUFFER_SIZE)
                uncompress = self._decompressor.decompress(buf, size)
            else:
                uncompress = self._decompressor.decompress(b"", size)
            if self._decompressor.unused_data != b"":
                # Prepend the already read bytes to the fileobj so they can
                # be seen by _read_eof() and _read_gzip_header()
                self._fp.prepend(self._decompressor.unused_data)

            if uncompress != b"":
                break
            if buf == b"":
                raise EOFError("Compressed file ended before the "
                               "end-of-stream marker was reached")

        self._add_read_data(uncompress)
        self._pos += len(uncompress)
        return uncompress


# Aliases for improved compatibility with CPython gzip module.
GzipFile = IGzipFile
_GzipReader = _IGzipReader


def _create_simple_gzip_header(compresslevel: int,
                               mtime: Optional[SupportsInt] = None) -> bytes:
    """
    Write a simple gzip header with no extra fields.
    :param compresslevel: Compresslevel used to determine the xfl bytes.
    :param mtime: The mtime (must support conversion to a 32-bit integer).
    :return: A bytes object representing the gzip header.
    """
    if mtime is None:
        mtime = time.time()
    # There is no best compression level. ISA-L only provides algorithms for
    # fast and medium levels.
    xfl = 4 if compresslevel == _COMPRESS_LEVEL_FAST else 0
    # Pack ID1 and ID2 magic bytes, method (8=deflate), header flags (no extra
    # fields added to header), mtime, xfl and os (255 for unknown OS).
    return struct.pack("<BBBBLBB", 0x1f, 0x8b, 8, 0, int(mtime), xfl, 255)


def compress(data, compresslevel=_COMPRESS_LEVEL_BEST, *, mtime=None):
    """Compress data in one shot and return the compressed string.
    Optional argument is the compression level, in range of 0-3.
    """
    header = _create_simple_gzip_header(compresslevel, mtime)
    # use igzip_lib to compress the data without a gzip header but with a
    # gzip trailer.
    compressed = igzip_lib.compress(data, compresslevel,
                                    flag=igzip_lib.COMP_GZIP_NO_HDR)
    return header + compressed


def _gzip_header_end(data: bytes) -> int:
    """
    Find the start of the raw deflate block in a gzip file.
    :param data: Compressed data that starts with a gzip header.
    :return: The end of the header / start of the raw deflate block.
    """
    eof_error = EOFError("Compressed file ended before the end-of-stream "
                         "marker was reached")
    if len(data) < 10:
        raise eof_error
    # We are not interested in mtime, xfl and os flags.
    magic, method, flags = struct.unpack("<HBB", data[:4])
    if magic != 0x8b1f:
        raise BadGzipFile(f"Not a gzipped file ({repr(data[:2])})")
    if method != 8:
        raise BadGzipFile("Unknown compression method")
    pos = 10
    if flags & FEXTRA:
        if len(data) < pos + 2:
            raise eof_error
        xlen = int.from_bytes(data[pos: pos + 2], "little", signed=False)
        pos += 2 + xlen
    if flags & FNAME:
        pos = data.find(b"\x00", pos) + 1
        # pos will be -1 + 1 when null byte not found.
        if not pos:
            raise eof_error
    if flags & FCOMMENT:
        pos = data.find(b"\x00", pos) + 1
        if not pos:
            raise eof_error
    if flags & FHCRC:
        if len(data) < pos + 2:
            raise eof_error
        header_crc = int.from_bytes(data[pos: pos + 2], "little", signed=False)
        # CRC is stored as a 16-bit integer by taking last bits of crc32.
        crc = isal_zlib.crc32(data[:pos]) & 0xFFFF
        if header_crc != crc:
            raise BadGzipFile(f"Corrupted header. Checksums do not "
                              f"match: {crc} != {header_crc}")
        pos += 2
    return pos


def decompress(data):
    """Decompress a gzip compressed string in one shot.
    Return the decompressed string.
    """
    all_blocks: List[bytes] = []
    while True:
        if data == b"":
            break
        header_end = _gzip_header_end(data)
        do = isal_zlib.decompressobj(-15)
        block = do.decompress(data[header_end:]) + do.flush()
        if not do.eof or len(do.unused_data) < 8:
            raise EOFError("Compressed file ended before the end-of-stream "
                           "marker was reached")
        checksum, length = struct.unpack("<II", do.unused_data[:8])
        crc = isal_zlib.crc32(block)
        if crc != checksum:
            raise BadGzipFile("CRC check failed")
        if length != len(block):
            raise BadGzipFile("Incorrect length of data produced")
        all_blocks.append(block)
        # Remove all padding null bytes and start next block.
        data = do.unused_data[8:].lstrip(b"\x00")
    return b"".join(all_blocks)


def _argument_parser():
    parser = argparse.ArgumentParser()
    parser.description = (
        "A simple command line interface for the igzip module. "
        "Acts like igzip.")
    parser.add_argument("file", nargs="?")
    compress_group = parser.add_mutually_exclusive_group()
    compress_group.add_argument(
        "-0", "--fast", action="store_const", dest="compresslevel",
        const=_COMPRESS_LEVEL_FAST,
        help="use compression level 0 (fastest)")
    compress_group.add_argument(
        "-1", action="store_const", dest="compresslevel",
        const=1,
        help="use compression level 1")
    compress_group.add_argument(
        "-2", action="store_const", dest="compresslevel",
        const=2,
        help="use compression level 2 (default)")
    compress_group.add_argument(
        "-3", "--best", action="store_const", dest="compresslevel",
        const=_COMPRESS_LEVEL_BEST,
        help="use compression level 3 (best)")
    compress_group.set_defaults(compress=True)
    compress_group.add_argument(
        "-d", "--decompress", action="store_const",
        dest="compress",
        const=False,
        help="Decompress the file instead of compressing.")
    output_group = parser.add_mutually_exclusive_group()
    output_group.add_argument("-c", "--stdout", action="store_true",
                              help="write on standard output")
    output_group.add_argument("-o", "--output",
                              help="Write to this output file")
    parser.add_argument("-f", "--force", action="store_true",
                        help="Overwrite output without prompting")
    # -b flag not taken by either gzip or igzip. Hidden attribute. Above 32K
    # diminishing returns hit. _compression.BUFFER_SIZE = 8k. But 32K is about
    # ~6% faster.
    parser.add_argument("-b", "--buffer-size",
                        default=128 * 1024, type=int,
                        help=argparse.SUPPRESS)
    return parser


def main():
    args = _argument_parser().parse_args()

    compresslevel = args.compresslevel or _COMPRESS_LEVEL_TRADEOFF

    if args.output:
        out_filepath = args.output
    elif args.stdout:
        out_filepath = None  # to stdout
    elif args.file is None:
        out_filepath = None  # to stout
    else:
        if args.compress:
            out_filepath = args.file + ".gz"
        else:
            out_filepath, extension = os.path.splitext(args.file)
            if extension != ".gz" and not args.stdout:
                sys.exit(f"filename doesn't end in .gz: {args.file!r}. "
                         f"Cannot determine output filename.")
    if out_filepath is not None and not args.force:
        if os.path.exists(out_filepath):
            yes_or_no = input(f"{out_filepath} already exists; "
                              f"do you wish to overwrite (y/n)?")
            if yes_or_no not in {"y", "Y", "yes"}:
                sys.exit("not overwritten")

    if args.compress:
        if args.file is None:
            in_file = sys.stdin.buffer
        else:
            in_file = io.open(args.file, mode="rb")
        if out_filepath is not None:
            out_file = open(out_filepath, "wb", compresslevel=compresslevel)
        else:
            out_file = IGzipFile(mode="wb", fileobj=sys.stdout.buffer,
                                 compresslevel=compresslevel)
    else:
        if args.file:
            in_file = open(args.file, mode="rb")
        else:
            in_file = IGzipFile(mode="rb", fileobj=sys.stdin.buffer)
        if out_filepath is not None:
            out_file = io.open(out_filepath, mode="wb")
        else:
            out_file = sys.stdout.buffer

    global READ_BUFFER_SIZE
    READ_BUFFER_SIZE = args.buffer_size
    try:
        while True:
            block = in_file.read(args.buffer_size)
            if block == b"":
                break
            out_file.write(block)
    finally:
        if in_file is not sys.stdin.buffer:
            in_file.close()
        if out_file is not sys.stdout.buffer:
            out_file.close()


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cython: language_level=3
# cython: binding=True

"""
Implementation of the zlib module using the ISA-L libraries.
"""
###############################################################################
###       README FOR DEVELOPERS: IMPLEMENTATION OF THIS MODULE              ###
###############################################################################
#
# This module is implemented with zlibmodule.c as example. Since there is a lot
# of duplication in zlibmodule.c there is a lot of duplication in this module.
# It is not always possible to route repetitive calls to a subroutine.
# Therefore the main methods are explained here.
#
# All compress, decompress and flush implementations are basically the same.
# 1. Get a buffer from the input data
# 2. Initialise an output buffer
# 3. Initialise a isal_zstream(for compression) or inflate_state (for
#    decompression). Hereafter referred as stream.
# 4. The amount of available input bytes is set on the stream. This is either
#    the maximum amount possible (in the case the input data is equal or larger
#    than the maximum amount). Or the length of the (remaining) input data.
# 5. The amount of available output bytes is set on the stream. Buffer is
#    enlarged as needed.
# 6. inflate/deflate/flush action
# 7. Check for errors in the action.
# 8. Was the output buffer completely filled? (stream.avail_out == 0). If so go
#    to 5. Special case: decompressobj. If the output buffer is at max_length
#    continue.
# 9. Was all the input read? if not go to 4. Alternatively in the case of
#    decompression: was the end of the stream reached? if not go to 4.
# 10. In case of decompression with leftover input data. For a decompressobj
#     this is stored in unconsumed_tail / unused_data.
# 11. Convert output buffer to bytes object and return.
#
# Errors are raised in the main functions as much as possible to prevent cdef
# functions returning PyObjects that need to be transformed in C variables.
# In cases where this is not possible, C variables were set using pointers.
# Allowing repeated use of functions while limiting the number of python
# interactions.
#
###############################################################################


import warnings
import zlib

from .crc cimport crc32_gzip_refl
# Import isa-l igzip-lib C constants and functions
from .igzip_lib cimport (
    ISAL_DEF_MAX_HIST_BITS, NO_FLUSH, SYNC_FLUSH, FULL_FLUSH, IGZIP_DEFLATE,
    IGZIP_GZIP, IGZIP_ZLIB, COMP_OK, ISAL_DECOMP_OK, ISAL_BLOCK_FINISH,
    ZSTATE_END, ISAL_DEFLATE, ISAL_GZIP, ISAL_ZLIB, ISAL_DEF_MIN_LEVEL,
    ISAL_DEF_MAX_LEVEL, isal_zstream, inflate_state, isal_deflate_init,
    isal_deflate_set_dict, isal_deflate, isal_inflate_init,
    isal_inflate_set_dict, isal_inflate, isal_adler32)
# Import python-isal igzip_lib cython functions
from .igzip_lib cimport(
    check_isal_deflate_rc, check_isal_inflate_rc,
    arrange_output_buffer_with_maximum, arrange_output_buffer,
    arrange_input_buffer, MEM_LEVEL_DEFAULT_I, MEM_LEVEL_MIN_I,
    MEM_LEVEL_SMALL_I, MEM_LEVEL_MEDIUM_I, MEM_LEVEL_LARGE_I,
    MEM_LEVEL_EXTRA_LARGE_I, ISAL_DEFAULT_COMPRESSION_I, mem_level_to_bufsize,
    view_bitbuffer,)

# Alias igzip_lib compress and decompress functions
from .igzip_lib cimport _compress as igzip_compress
from .igzip_lib cimport _decompress as igzip_decompress

from . import igzip_lib
from libc.stdint cimport UINT64_MAX, UINT32_MAX
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
from cpython.buffer cimport PyBUF_C_CONTIGUOUS, PyObject_GetBuffer, PyBuffer_Release
from cpython.bytes cimport PyBytes_FromStringAndSize
from cpython.long cimport PyLong_AsUnsignedLongMask

cdef extern from "<Python.h>":
    const Py_ssize_t PY_SSIZE_T_MAX

ISAL_BEST_SPEED = igzip_lib.ISAL_BEST_SPEED
ISAL_BEST_COMPRESSION = igzip_lib.ISAL_BEST_COMPRESSION
ISAL_DEFAULT_COMPRESSION = igzip_lib.ISAL_DEFAULT_COMPRESSION
Z_BEST_SPEED = ISAL_BEST_SPEED
Z_BEST_COMPRESSION = ISAL_BEST_COMPRESSION
Z_DEFAULT_COMPRESSION = ISAL_DEFAULT_COMPRESSION

# Compile time constants with _I (for integer suffix) as names without
# suffix should be exposed to the user.
DEF DEF_BUF_SIZE_I = 16 * 1024
DEF DEF_MEM_LEVEL_I = 8

# Expose compile-time constants. Same names as zlib.
DEF_BUF_SIZE = DEF_BUF_SIZE_I
DEF_MEM_LEVEL = DEF_MEM_LEVEL_I
MAX_WBITS = igzip_lib.MAX_HIST_BITS

# Compression methods
DEFLATED = zlib.DEFLATED

# Strategies
Z_DEFAULT_STRATEGY=zlib.Z_DEFAULT_STRATEGY
Z_HUFFMAN_ONLY=zlib.Z_HUFFMAN_ONLY
Z_FILTERED=zlib.Z_FILTERED
# Following strategies not supported on all versions of zlib.
if hasattr(zlib, "Z_RLE"):
    Z_RLE = zlib.Z_RLE
if hasattr(zlib, "Z_FIXED"):
   Z_FIXED=zlib.Z_FIXED

# Flush methods
Z_NO_FLUSH=zlib.Z_NO_FLUSH
Z_SYNC_FLUSH=zlib.Z_SYNC_FLUSH
Z_FULL_FLUSH=zlib.Z_FULL_FLUSH
Z_FINISH=zlib.Z_FINISH

# Add error for compatibility
IsalError = igzip_lib.IsalError
error = IsalError


if ISAL_DEF_MAX_HIST_BITS > zlib.MAX_WBITS:
    raise  IsalError("ISAL max window size no longer compatible with zlib. "
                     "Please contact the developers.")


def adler32(data, value = 1):
    """
    Computes an Adler-32 checksum of *data*. Returns the checksum as unsigned
    32-bit integer.

    :param data: Binary data (bytes, bytearray, memoryview).
    :param value: The starting value of the checksum.
    """
    cdef unsigned long init = PyLong_AsUnsignedLongMask(value)
    cdef Py_buffer buffer_data
    cdef Py_buffer* buffer = &buffer_data
    # Cython makes sure error is handled when acquiring buffer fails.
    PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
    try:
        if buffer.len > UINT64_MAX:
            raise ValueError("Data too big for adler32")
        return isal_adler32(init, <unsigned char*>buffer.buf, buffer.len)
    finally:
        PyBuffer_Release(buffer)

def crc32(data, value = 0):
    """
    Computes a CRC-32 checksum of *data*. Returns the checksum as unsigned
    32-bit integer.

    :param data: Binary data (bytes, bytearray, memoryview).
    :param value: The starting value of the checksum.
    """
    cdef unsigned long init = PyLong_AsUnsignedLongMask(value)
    cdef Py_buffer buffer_data
    cdef Py_buffer* buffer = &buffer_data
    # Cython makes sure error is handled when acquiring buffer fails.
    PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
    try:
        if buffer.len > UINT64_MAX:
            raise ValueError("Data too big for adler32")
        return crc32_gzip_refl(init, <unsigned char*>buffer.buf, buffer.len)
    finally:
        PyBuffer_Release(buffer)


def compress(data,
             int level=ISAL_DEFAULT_COMPRESSION_I,
             int wbits = ISAL_DEF_MAX_HIST_BITS):
    """
    Compresses the bytes in *data*. Returns a bytes object with the
    compressed data.

    :param level: the compression level from 0 to 3. 0 is the lowest
                  compression (NOT no compression as in stdlib zlib!) and the
                  fastest. 3 is the best compression and the slowest. Default
                  is a compromise at level 2.

    :param wbits: Set the amount of history bits or window size and which
                  headers and trailers are used. Values from 9 to 15 signify
                  will use a zlib header and trailer. From +25 to +31
                  (16 + 9 to 15) a gzip header and trailer will be used.
                  -9 to -15 will generate a raw compressed string with
                  no headers and trailers.
    """
    cdef unsigned short hist_bits
    cdef unsigned short flag
    wbits_to_flag_and_hist_bits_deflate(wbits,
                                        &hist_bits,
                                        &flag)
    return igzip_compress(data, level, flag, MEM_LEVEL_DEFAULT_I, hist_bits)

def decompress(data,
                 int wbits=ISAL_DEF_MAX_HIST_BITS,
                 Py_ssize_t bufsize=DEF_BUF_SIZE,):
    """
    Deompresses the bytes in *data*. Returns a bytes object with the
    decompressed data.
    
    :param wbits: Set the amount of history bits or window size and which
                  headers and trailers are expected. Values from 8 to 15
                  will expect a zlib header and trailer. -8 to -15 will expect
                  a raw compressed string with no headers and trailers.
                  From +24 to +31 == 16 + (8 to 15) a gzip header and trailer
                  will be expected. From +40 to +47 == 32 + (8 to 15)
                  automatically detects a gzip or zlib header.
    :param bufsize: The initial size of the output buffer.
    """
    cdef unsigned int hist_bits
    cdef unsigned int flag
    cdef bint is_gzip
    data_is_gzip(data, &is_gzip)
    wbits_to_flag_and_hist_bits_inflate(wbits,
                                        &hist_bits,
                                        &flag,
                                        is_gzip)
    return igzip_decompress(data, flag, hist_bits, bufsize)


def decompressobj(int wbits=ISAL_DEF_MAX_HIST_BITS,
                  zdict = None):
    """
    Returns a Decompress object for decompressing data streams.

    :param wbits: Set the amount of history bits or window size and which
                  headers and trailers are expected. Values from 8 to 15
                  will expect a zlib header and trailer. -8 to -15 will expect
                  a raw compressed string with no headers and trailers.
                  From +24 to +31 == 16 + (8 to 15) a gzip header and trailer
                  will be expected. From +40 to +47 == 32 + (8 to 15)
                  automatically detects a gzip or zlib header.
    :zdict:       A predefined compression dictionary. Must be the same zdict
                  as was used to compress the data.
    """
    return Decompress.__new__(Decompress, wbits, zdict)


def compressobj(int level=ISAL_DEFAULT_COMPRESSION_I,
                int method=DEFLATED,
                int wbits=ISAL_DEF_MAX_HIST_BITS,
                int memLevel=DEF_MEM_LEVEL,
                int strategy=zlib.Z_DEFAULT_STRATEGY,
                zdict = None):
    """
    Returns a Compress object for compressing data streams.

    :param level:   the compression level from 0 to 3. 0 is the lowest
                    compression (NOT no compression as in stdlib zlib!) and the
                    fastest. 3 is the best compression and the slowest. Default
                    is a compromise at level 2.
    :param method:  The compression algorithm. Currently only DEFLATED is
                    supported.
    :param wbits:   Set the amount of history bits or window size and which
                    headers and trailers are used. Values from 9 to 15 signify
                    will use a zlib header and trailer. From +25 to +31
                    (16 + 9 to 15) a gzip header and trailer will be used.
                    -9 to -15 will generate a raw compressed string with
                    no headers and trailers.
    :param memLevel: The amount of memory used for the internal compression
                     state. Higher values use more memory for better speed and
                     smaller output. Values between 1 and 9 are supported.
    :zdict:         A predefined compression dictionary. A sequence of bytes
                    that are expected to occur frequently in the to be
                    compressed data. The most common subsequences should come
                    at the end.
    """
    return Compress.__new__(Compress, level, method, wbits, memLevel, strategy, zdict)


cdef class Compress:
    """Compress object for handling streaming compression."""
    cdef isal_zstream stream
    cdef unsigned char * level_buf

    def __cinit__(self,
                  int level = ISAL_DEFAULT_COMPRESSION_I,
                  int method = DEFLATED,
                  int wbits = ISAL_DEF_MAX_HIST_BITS,
                  int memLevel = DEF_MEM_LEVEL,
                  int strategy = Z_DEFAULT_STRATEGY,
                  zdict = None):
        if strategy != Z_DEFAULT_STRATEGY:
            warnings.warn("Only one strategy is supported when using "
                          "isal_zlib. Using the default strategy.")
        isal_deflate_init(&self.stream)

        wbits_to_flag_and_hist_bits_deflate(wbits,
                                            &self.stream.hist_bits,
                                            &self.stream.gzip_flag)

        cdef Py_ssize_t zdict_length
        if zdict:
            zdict_length = len(zdict)
            if zdict_length > UINT32_MAX:
                raise OverflowError("zdict length does not fit in an unsigned int")
            err = isal_deflate_set_dict(&self.stream, zdict, zdict_length)
            if err != COMP_OK:
                check_isal_deflate_rc(err)
        self.stream.level = level
        zlib_mem_level_to_isal_bufsize(level, memLevel, &self.stream.level_buf_size)
        self.level_buf = <unsigned char *>PyMem_Malloc(self.stream.level_buf_size * sizeof(char))
        self.stream.level_buf = self.level_buf

    def __dealloc__(self):
        if self.level_buf is not NULL:
            PyMem_Free(self.level_buf)

    def compress(self, data):
        """
        Compress *data* returning a bytes object with at least part of the
        data in *data*. This data should be concatenated to the output
        produced by any preceding calls to the compress() method.
        Some input may be kept in internal buffers for later processing.
        """
        # Initialise output buffer
        cdef unsigned char * obuf = NULL
        cdef Py_ssize_t obuflen = DEF_BUF_SIZE_I

        # initialise input
        cdef Py_buffer buffer_data
        cdef Py_buffer* buffer = &buffer_data
        # Cython makes sure error is handled when acquiring buffer fails.
        PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
        cdef Py_ssize_t ibuflen = buffer.len
        self.stream.next_in = <unsigned char*>buffer.buf

        # initialise helper variables
        cdef int err
        try:
            while True:
                arrange_input_buffer(&self.stream, &ibuflen)
                while True:
                    obuflen = arrange_output_buffer(&self.stream, &obuf, obuflen)
                    if obuflen== -1:
                        raise MemoryError("Unsufficient memory for buffer allocation")
                    err = isal_deflate(&self.stream)
                    if err != COMP_OK:
                        check_isal_deflate_rc(err)
                    if self.stream.avail_out != 0:
                        break
                if self.stream.avail_in != 0:
                    raise AssertionError("Input stream should be empty")
                if ibuflen == 0:
                    break
            return PyBytes_FromStringAndSize(<char*>obuf, self.stream.next_out - obuf)
        finally:
            PyBuffer_Release(buffer)
            PyMem_Free(obuf)

    def flush(self, mode=zlib.Z_FINISH):
        """
        All pending input is processed, and a bytes object containing the
        remaining compressed output is returned.

        :param mode: Defaults to Z_FINISH which
                     finishes the compressed stream and prevents compressing
                     any more data. The other supported methods are
                     Z_NO_FLUSH, Z_SYNC_FLUSH and Z_FULL_FLUSH.
        """

        if mode == zlib.Z_NO_FLUSH:
            # Flushing with no_flush does nothing.
            return b""
        elif mode == zlib.Z_FINISH:
            self.stream.flush = FULL_FLUSH
            self.stream.end_of_stream = 1
        elif mode == zlib.Z_FULL_FLUSH:
            self.stream.flush = FULL_FLUSH
        elif mode == zlib.Z_SYNC_FLUSH:
            self.stream.flush=SYNC_FLUSH
        else:
            raise IsalError("Unsupported flush mode")

        cdef Py_ssize_t length = DEF_BUF_SIZE_I
        cdef unsigned char * obuf = NULL

        try:
            while True:
                length = arrange_output_buffer(&self.stream, &obuf, length)
                if length == -1:
                    raise MemoryError("Unsufficient memory for buffer allocation")
                err = isal_deflate(&self.stream)
                if err != COMP_OK:
                    check_isal_deflate_rc(err)
                if self.stream.avail_out != 0:
                    break
            if self.stream.avail_in != 0:
                raise AssertionError("There should be no available input after flushing.")
            return PyBytes_FromStringAndSize(<char*>obuf, self.stream.next_out - obuf)
        finally:
            PyMem_Free(obuf)

cdef class Decompress:
    """Decompress object for handling streaming decompression."""
    cdef public bytes unused_data
    cdef public bytes unconsumed_tail
    cdef public bint eof
    cdef inflate_state stream
    cdef bint method_set
    cdef bint is_gzip

    def __cinit__(self, int wbits=ISAL_DEF_MAX_HIST_BITS, zdict = None):
        isal_inflate_init(&self.stream)

        wbits_to_flag_and_hist_bits_inflate(wbits,
                                            &self.stream.hist_bits,
                                            &self.stream.crc_flag)
        if 40 <= wbits <= 47:
            self.method_set = 0
        else:
            self.method_set = 1

        cdef Py_ssize_t zdict_length
        if zdict:
            zdict_length = len(zdict)
            if zdict_length > UINT32_MAX:
                raise OverflowError("zdict length does not fit in an unsigned int")
            err = isal_inflate_set_dict(&self.stream, zdict, zdict_length)
            if err != COMP_OK:
                check_isal_deflate_rc(err)
        self.unused_data = b""
        self.unconsumed_tail = b""
        self.eof = False

    def _view_bitbuffer(self):
        """Shows the 64-bitbuffer of the internal inflate_state. It contains
        a maximum of 8 bytes. This data is already read-in so is not part
        of the unconsumed tail."""
        return view_bitbuffer(&self.stream)

    cdef save_unconsumed_input(self, Py_buffer *data):
        cdef Py_ssize_t old_size, new_size, left_size
        cdef bytes new_data
        if self.stream.block_state == ISAL_BLOCK_FINISH:
            self.eof = 1
            if self.stream.avail_in > 0:
                left_size = <unsigned char*>data.buf + data.len - self.stream.next_in
                new_data = PyBytes_FromStringAndSize(<char *>self.stream.next_in, left_size)
            else:
                new_data = b""
            if not self.unused_data:
                # The block is finished and this decompressobject can not be
                # used anymore. Some unused data is in the bitbuffer and has to
                # be recovered. Only when self.unused_data is empty. Otherwise
                # we assume the bitbuffer data is already added.
                self.unused_data = self._view_bitbuffer()
            self.unused_data += new_data
            if self.unconsumed_tail:
                self.unconsumed_tail = b""  # When there is unused_data unconsumed tail should be b""
        elif self.stream.avail_in > 0 or self.unconsumed_tail:
            left_size = <unsigned char*>data.buf + data.len - self.stream.next_in
            new_data = PyBytes_FromStringAndSize(<char *>self.stream.next_in, left_size)
            self.unconsumed_tail = new_data

    def decompress(self, data, Py_ssize_t max_length = 0):
        """
        Decompress data, returning a bytes object containing the uncompressed
        data corresponding to at least part of the data in string.

        :param max_length: if non-zero then the return value will be no longer
                           than max_length. Unprocessed data will be in the
                           unconsumed_tail attribute.
        """
   
        cdef Py_ssize_t hard_limit
        if max_length == 0:
            hard_limit = PY_SSIZE_T_MAX
        elif max_length < 0:
            raise ValueError("max_length can not be smaller than 0")
        else:
            hard_limit = max_length

        if not self.method_set:
            # Try to detect method from the first two bytes of the data.
            data_is_gzip(data, &self.is_gzip)
            self.stream.crc_flag = ISAL_GZIP if self.is_gzip else ISAL_ZLIB
            self.method_set = 1

        # initialise input
        cdef Py_buffer buffer_data
        cdef Py_buffer* buffer = &buffer_data
        # Cython makes sure error is handled when acquiring buffer fails.
        PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
        cdef Py_ssize_t ibuflen = buffer.len
        self.stream.next_in = <unsigned char*>buffer.buf

        cdef int err
        cdef bint max_length_reached = False
        
        # Initialise output buffer
        cdef unsigned char *obuf = NULL
        cdef Py_ssize_t obuflen = DEF_BUF_SIZE_I
        if obuflen > hard_limit:
            obuflen = hard_limit

        try:
            while True:
                arrange_input_buffer(&self.stream, &ibuflen)
                while True:
                    obuflen = arrange_output_buffer_with_maximum(
                              &self.stream, &obuf, obuflen, hard_limit)
                    if obuflen == -1:
                        raise MemoryError("Unsufficient memory for buffer allocation")
                    elif obuflen == -2:
                        max_length_reached = True
                        break
                    err = isal_inflate(&self.stream)
                    if err != ISAL_DECOMP_OK:
                        check_isal_inflate_rc(err)
                    if self.stream.block_state == ISAL_BLOCK_FINISH or self.stream.avail_out != 0:
                        break
                if self.stream.block_state == ISAL_BLOCK_FINISH or ibuflen ==0 or max_length_reached:
                    break
            self.save_unconsumed_input(buffer)
            return PyBytes_FromStringAndSize(<char*>obuf, self.stream.next_out - obuf)
        finally:
            PyBuffer_Release(buffer)
            PyMem_Free(obuf)

    def flush(self, Py_ssize_t length = DEF_BUF_SIZE):
        """
        All pending input is processed, and a bytes object containing the
        remaining uncompressed output is returned.

        :param length: The initial size of the output buffer.
        """
        if length <= 0:
            raise ValueError("Length must be greater than 0")

        cdef Py_buffer buffer_data
        cdef Py_buffer* buffer = &buffer_data
        # Cython makes sure error is handled when acquiring buffer fails.
        PyObject_GetBuffer(self.unconsumed_tail, buffer, PyBUF_C_CONTIGUOUS)
        cdef Py_ssize_t ibuflen = buffer.len
        self.stream.next_in = <unsigned char*>buffer.buf

        cdef unsigned int obuflen = length
        cdef unsigned char * obuf = NULL

        cdef int err

        try:
            while True:
                arrange_input_buffer(&self.stream, &ibuflen)
                while True:
                    obuflen = arrange_output_buffer(&self.stream, &obuf, obuflen)
                    if obuflen == -1:
                        raise MemoryError("Unsufficient memory for buffer allocation")
                    err = isal_inflate(&self.stream)
                    if err != ISAL_DECOMP_OK:
                        check_isal_inflate_rc(err)
                    if self.stream.avail_out != 0 or self.stream.block_state == ISAL_BLOCK_FINISH:
                        break
                if self.stream.block_state == ISAL_BLOCK_FINISH or ibuflen == 0:
                    break
            self.save_unconsumed_input(buffer)
            return PyBytes_FromStringAndSize(<char*>obuf, self.stream.next_out - obuf)
        finally:
            PyBuffer_Release(buffer)
            PyMem_Free(obuf)

cdef data_is_gzip(object data, bint *is_gzip):
    cdef Py_buffer buffer_data
    cdef Py_buffer* buffer = &buffer_data
    PyObject_GetBuffer(data, buffer, PyBUF_C_CONTIGUOUS)
    if buffer.len < 2:
        PyBuffer_Release(buffer)
        is_gzip[0] = False
        return
    cdef unsigned char * char_ptr = <unsigned char *>buffer.buf
    if char_ptr[0] == 31:
        if char_ptr[1] == 139:
            PyBuffer_Release(buffer)
            is_gzip[0] = True
            return
    PyBuffer_Release(buffer)
    is_gzip[0] = False
    return


cdef wbits_to_flag_and_hist_bits_deflate(int wbits,
                                         unsigned short * hist_bits,
                                         unsigned short * gzip_flag):
    if 9 <= wbits <= 15:  # zlib headers and trailers on compressed stream
        hist_bits[0] = wbits
        gzip_flag[0] = IGZIP_ZLIB
    elif 25 <= wbits <= 31:  # gzip headers and trailers on compressed stream
        hist_bits[0] = wbits - 16
        gzip_flag[0] = IGZIP_GZIP
    elif -15 <= wbits <= -9:  # raw compressed stream
        hist_bits[0] = -wbits
        gzip_flag[0] = IGZIP_DEFLATE
    else:
        raise ValueError("Invalid wbits value")


cdef wbits_to_flag_and_hist_bits_inflate(int wbits,
                                         unsigned int * hist_bits,
                                         unsigned int * crc_flag,
                                         bint gzip = 0):
    if wbits == 0:
        hist_bits[0] = 0
        crc_flag[0] = ISAL_ZLIB
    elif 8 <= wbits <= 15:  # zlib headers and trailers on compressed stream
        hist_bits[0] = wbits
        crc_flag[0] = ISAL_ZLIB
    elif 24 <= wbits <= 31:  # gzip headers and trailers on compressed stream
        hist_bits[0] = wbits - 16
        crc_flag[0] = ISAL_GZIP
    elif -15 <= wbits <= -8:  # raw compressed stream
        hist_bits[0] = -wbits
        crc_flag[0] = ISAL_DEFLATE
    elif 40 <= wbits <= 47:  # Accept gzip or zlib
        hist_bits[0] = wbits - 32
        crc_flag[0] = ISAL_GZIP if gzip else ISAL_ZLIB
    else:
        raise ValueError("Invalid wbits value")

cdef int zlib_mem_level_to_isal_bufsize(int compression_level, int mem_level, unsigned int *bufsize):
    """
    Convert zlib memory levels to isal equivalents
    """
    if not (1 <= mem_level <= 9):
        bufsize[0] = 0
        return -1
    if not (ISAL_DEF_MIN_LEVEL <= compression_level <= ISAL_DEF_MAX_LEVEL):
        bufsize[0] = 0
        return -1

    # If the mem_level is zlib default, return isal defaults.
    # Current zlib def level = 8. On isal the def level is large.
    # Hence 7,8 return large. 9 returns extra large.
    cdef int level = MEM_LEVEL_DEFAULT_I
    if mem_level == DEF_MEM_LEVEL_I:
        level = MEM_LEVEL_DEFAULT_I
    elif mem_level == 1:
        level = MEM_LEVEL_MIN_I
    elif mem_level in [2,3]:
        level = MEM_LEVEL_SMALL_I
    elif mem_level in [4,5,6]:
        level = MEM_LEVEL_MEDIUM_I
    elif mem_level in [7,8]:
        level = MEM_LEVEL_LARGE_I
    elif mem_level == 9:
        level = MEM_LEVEL_EXTRA_LARGE_I
    else:
        bufsize[0] = 0
        return -1
    if mem_level_to_bufsize(compression_level, level, bufsize) != 0:
        bufsize[0] = 0
        return -1
    return 0


# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import functools
import os
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

ISA_L_SOURCE = os.path.join("src", "isal", "isa-l")

SYSTEM_IS_UNIX = (sys.platform.startswith("linux") or
                  sys.platform.startswith("darwin"))
SYSTEM_IS_WINDOWS = sys.platform.startswith("win")


class IsalExtension(Extension):
    """Custom extension to allow for targeted modification."""
    pass


MODULES = [IsalExtension("isal.isal_zlib", ["src/isal/isal_zlib.pyx"]),
           IsalExtension("isal.igzip_lib", ["src/isal/igzip_lib.pyx"])]
if SYSTEM_IS_UNIX:
    MODULES.append(IsalExtension("isal._isal", ["src/isal/_isal.pyx"]))


class BuildIsalExt(build_ext):
    def build_extension(self, ext):
        if not isinstance(ext, IsalExtension):
            super().build_extension(ext)
            return

        # Add option to link dynamically for packaging systems such as conda.
        # Always link dynamically on readthedocs to simplify install.
        if (os.getenv("PYTHON_ISAL_LINK_DYNAMIC") is not None or
                os.environ.get("READTHEDOCS") is not None):
            # Check for isa-l include directories. This is useful when
            # installing in a conda environment.
            possible_prefixes = [sys.exec_prefix, sys.base_exec_prefix]
            for prefix in possible_prefixes:
                if Path(prefix, "include", "isa-l").exists():
                    ext.include_dirs = [os.path.join(prefix, "include")]
                    ext.library_dirs = [os.path.join(prefix, "lib")]
                    break   # Only one include directory is needed.
                # On windows include is in Library apparently
                elif Path(prefix, "Library", "include", "isa-l").exists():
                    ext.include_dirs = [os.path.join(prefix, "Library",
                                                     "include")]
                    ext.library_dirs = [os.path.join(prefix, "Library", "lib")]
                    break
            if SYSTEM_IS_UNIX:
                ext.libraries = ["isal"]  # libisal.so*
            elif SYSTEM_IS_WINDOWS:
                ext.libraries = ["isa-l"]  # isa-l.dll
            else:
                raise NotImplementedError(
                    f"Unsupported platform: {sys.platform}")
        else:
            if self.compiler.compiler_type == "msvc":
                compiler = copy.deepcopy(self.compiler)
                if not compiler.initialized:
                    compiler.initialize()
                compiler_command = f'"{compiler.cc}"'
                compiler_args = compiler.compile_options
            elif self.compiler.compiler_type == "unix":
                compiler_command = self.compiler.compiler[0]
                compiler_args = self.compiler.compiler[1:]
            else:
                raise NotImplementedError("Unknown compiler")
            isa_l_prefix_dir = build_isa_l(compiler_command,
                                           " ".join(compiler_args))
            if SYSTEM_IS_UNIX:
                ext.extra_objects = [
                    os.path.join(isa_l_prefix_dir, "lib", "libisal.a")]
            elif SYSTEM_IS_WINDOWS:
                ext.extra_objects = [
                    os.path.join(isa_l_prefix_dir, "isa-l_static.lib")]
            else:
                raise NotImplementedError(
                    f"Unsupported platform: {sys.platform}")
            ext.include_dirs = [os.path.join(isa_l_prefix_dir,
                                             "include")]
            # -fPIC needed for proper static linking
            ext.extra_compile_args = ["-fPIC"]
        if os.getenv("CYTHON_COVERAGE") is not None:
            # Import cython here so python setup.py can be used without
            # installing cython.
            from Cython.Build import cythonize
            # Add cython directives and macros for coverage support.
            cythonized_exts = cythonize(ext, compiler_directives=dict(
                linetrace=True
            ))
            for cython_ext in cythonized_exts:
                cython_ext.define_macros = [("CYTHON_TRACE_NOGIL", "1")]
                cython_ext._needs_stub = False
                super().build_extension(cython_ext)
            return
        super().build_extension(ext)


# Use a cache to prevent isa-l from being build twice. According to the
# functools docs lru_cache with maxsize None is faster. The shortcut called
# 'cache' is only available from python 3.9 onwards.
# see: https://docs.python.org/3/library/functools.html#functools.cache
@functools.lru_cache(maxsize=None)
def build_isa_l(compiler_command: str, compiler_options: str):
    # Creating temporary directories
    build_dir = tempfile.mktemp()
    temp_prefix = tempfile.mkdtemp()
    shutil.copytree(ISA_L_SOURCE, build_dir)

    # Build environment is a copy of OS environment to allow user to influence
    # it.
    build_env = os.environ.copy()
    # Add -fPIC flag to allow static compilation
    build_env["CC"] = compiler_command
    if SYSTEM_IS_UNIX:
        build_env["CFLAGS"] = compiler_options + " -fPIC"
    elif SYSTEM_IS_WINDOWS:
        # The nmake file has CLFAGS_REL for all the compiler options.
        # This is added to CFLAGS with all the necessary include options.
        build_env["CFLAGS_REL"] = compiler_options
    if hasattr(os, "sched_getaffinity"):
        cpu_count = len(os.sched_getaffinity(0))
    else:  # sched_getaffinity not available on all platforms
        cpu_count = os.cpu_count() or 1  # os.cpu_count() can return None
    run_args = dict(cwd=build_dir, env=build_env)
    if SYSTEM_IS_UNIX:
        subprocess.run(os.path.join(build_dir, "autogen.sh"), **run_args)
        subprocess.run([os.path.join(build_dir, "configure"),
                        "--prefix", temp_prefix], **run_args)
        subprocess.run(["make", "-j", str(cpu_count)], **run_args)
        subprocess.run(["make", "-j", str(cpu_count), "install"], **run_args)
    elif SYSTEM_IS_WINDOWS:
        subprocess.run(["nmake", "/E", "/f", "Makefile.nmake"], **run_args)
        Path(temp_prefix, "include").mkdir()
        print(temp_prefix, file=sys.stderr)
        shutil.copytree(os.path.join(build_dir, "include"),
                        Path(temp_prefix, "include", "isa-l"))
        shutil.copy(os.path.join(build_dir, "isa-l_static.lib"),
                    os.path.join(temp_prefix, "isa-l_static.lib"))
    else:
        raise NotImplementedError(f"Unsupported platform: {sys.platform}")
    shutil.rmtree(build_dir)
    return temp_prefix


setup(
    name="isal",
    version="0.11.1",
    description="Faster zlib and gzip compatible compression and "
                "decompression by providing python bindings for the ISA-L "
                "library.",
    author="Leiden University Medical Center",
    author_email="r.h.p.vorderman@lumc.nl",  # A placeholder for now
    long_description=Path("README.rst").read_text(),
    long_description_content_type="text/x-rst",
    cmdclass={"build_ext": BuildIsalExt},
    license="MIT",
    keywords="isal isa-l compression deflate gzip igzip",
    zip_safe=False,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'isal': ['*.pxd', '*.pyx', '*.pyi', 'py.typed',
                           # Include isa-l LICENSE and other relevant files
                           # with the binary distribution.
                           'isa-l/LICENSE', 'isa-l/README.md',
                           'isa-l/Release_notes.txt']},
    url="https://github.com/pycompression/python-isal",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Cython",
        "Development Status :: 4 - Beta",
        "Topic :: System :: Archiving :: Compression",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.6",
    ext_modules=MODULES
)



# Copyright (c) 2020 Leiden University Medical Center
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

ISAL_BEST_SPEED: int
ISAL_BEST_COMPRESSION: int
ISAL_DEFAULT_COMPRESSION: int
Z_BEST_SPEED: int
Z_BEST_COMPRESSION: int
Z_DEFAULT_COMPRESSION: int

DEF_BUF_SIZE: int
DEF_MEM_LEVEL: int
MAX_WBITS: int

DEFLATED: int

Z_DEFAULT_STRATEGY: int
Z_RLE: int
Z_HUFFMAN_ONLY: int
Z_FILTERED: int
Z_FIXED: int

Z_NO_FLUSH: int
Z_SYNC_FLUSH: int
Z_FULL_FLUSH: int
Z_FINISH: int

class IsalError(OSError): ...

error: IsalError

def adler32(data, value: int = 1) -> int: ...
def crc32(data, value: int = 0) -> int: ...

def compress(data, level: int = ISAL_DEFAULT_COMPRESSION,
             wbits: int = MAX_WBITS) -> bytes: ...
def decompress(data, wbits: int = MAX_WBITS,
               bufsize: int = DEF_BUF_SIZE) -> bytes: ...

class Compress:
    def compress(self, data) -> bytes: ...
    def flush(self, mode: int = Z_FINISH) -> bytes: ...

class Decompress:
    unused_data: bytes
    unconsumed_tail: bytes
    eof: bool

    def decompress(self, data, max_length: int = 0) -> bytes: ...
    def flush(self, length: int = DEF_BUF_SIZE) -> bytes: ...

def compressobj(level: int = ISAL_DEFAULT_COMPRESSION,
                method: int = DEFLATED,
                wbits: int = MAX_WBITS,
                memLevel: int = DEF_MEM_LEVEL,
                strategy: int = Z_DEFAULT_STRATEGY,
                zdict = None) -> Compress: ...
def decompressobj(wbits: int = MAX_WBITS, zdict = None) -> Decompress: ...
